@@ -55,6 +55,8 @@ static cfg_opt_t sec_general[] =
     CFG_BOOL("ipv6", cfg_true, CFGF_NONE),
     CFG_STR("cache_path", STATEDIR "/cache/" PACKAGE "/cache.db", CFGF_NONE),
     CFG_INT("cache_daap_threshold", 1000, CFGF_NONE),
+    CFG_BOOL("speaker_autoselect", cfg_true, CFGF_NONE),
+    CFG_STR("allow_origin", "*", CFGF_NONE),
     CFG_END()
   };
 
@@ -84,6 +86,7 @@ static cfg_opt_t sec_library[] =
     CFG_STR_LIST("filepath_ignore", NULL, CFGF_NONE),
     CFG_BOOL("filescan_disable", cfg_false, CFGF_NONE),
     CFG_BOOL("itunes_overrides", cfg_false, CFGF_NONE),
+    CFG_BOOL("itunes_smartpl", cfg_false, CFGF_NONE),
     CFG_STR_LIST("no_decode", NULL, CFGF_NONE),
     CFG_STR_LIST("force_decode", NULL, CFGF_NONE),
     CFG_END()
@@ -96,6 +99,7 @@ static cfg_opt_t sec_audio[] =
     CFG_STR("type", NULL, CFGF_NONE),
     CFG_STR("card", "default", CFGF_NONE),
     CFG_STR("mixer", NULL, CFGF_NONE),
+    CFG_STR("mixer_device", NULL, CFGF_NONE),
     CFG_INT("offset", 0, CFGF_NONE),
     CFG_END()
   };
@@ -108,6 +112,14 @@ static cfg_opt_t sec_airplay[] =
     CFG_END()
   };
 
+/* FIFO section structure */
+static cfg_opt_t sec_fifo[] =
+  {
+    CFG_STR("nickname", "fifo", CFGF_NONE),
+    CFG_STR("path", NULL, CFGF_NONE),
+    CFG_END()
+  };
+
 /* Spotify section structure */
 static cfg_opt_t sec_spotify[] =
   {
@@ -147,6 +159,7 @@ static cfg_opt_t toplvl_cfg[] =
     CFG_SEC("library", sec_library, CFGF_NONE),
     CFG_SEC("audio", sec_audio, CFGF_NONE),
     CFG_SEC("airplay", sec_airplay, CFGF_MULTI | CFGF_TITLE),
+    CFG_SEC("fifo", sec_fifo, CFGF_NONE),
     CFG_SEC("spotify", sec_spotify, CFGF_NONE),
     CFG_SEC("sqlite", sec_sqlite, CFGF_NONE),
     CFG_SEC("mpd", sec_mpd, CFGF_NONE),
