@@ -23,6 +23,7 @@
 # include <config.h>
 #endif
 
+#include <libgen.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -33,32 +34,24 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <sys/inotify.h>
 #include <fcntl.h>
 #include <dirent.h>
 #include <pthread.h>
+#ifdef HAVE_PTHREAD_NP_H
+# include <pthread_np.h>
+#endif
 
 #include <unistr.h>
 #include <unictype.h>
 #include <uninorm.h>
 
-#if defined(__linux__)
-# include <sys/inotify.h>
-#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-# include <sys/time.h>
-# include <sys/event.h>
-#endif
-
-#if defined(HAVE_SYS_EVENTFD_H) && defined(HAVE_EVENTFD)
-# define USE_EVENTFD
-# include <sys/eventfd.h>
-#endif
+#include <event2/event.h>
 
 #ifdef HAVE_REGEX_H
 # include <regex.h>
 #endif
 
-#include <event.h>
-
 #include "logger.h"
 #include "db.h"
 #include "filescanner.h"
@@ -68,6 +61,7 @@
 #include "player.h"
 #include "cache.h"
 #include "artwork.h"
+#include "commands.h"
 
 #ifdef LASTFM
 # include "lastfm.h"
@@ -77,22 +71,6 @@
 #endif
 
 
-struct filescanner_command;
-
-typedef int (*cmd_func)(struct filescanner_command *cmd);
-
-struct filescanner_command
-{
-  pthread_mutex_t lck;
-  pthread_cond_t cond;
-
-  cmd_func func;
-
-  int nonblock;
-
-  int ret;
-};
-
 #define F_SCAN_BULK    (1 << 0)
 #define F_SCAN_RESCAN  (1 << 1)
 #define F_SCAN_FAST    (1 << 2)
@@ -117,29 +95,37 @@ struct deferred_pl {
   char *path;
   time_t mtime;
   struct deferred_pl *next;
+  int directory_id;
 };
 
 struct stacked_dir {
   char *path;
+  int parent_id;
   struct stacked_dir *next;
 };
 
-
-static int cmd_pipe[2];
-#ifdef USE_EVENTFD
-static int exit_efd;
-#else
-static int exit_pipe[2];
-#endif
 static int scan_exit;
 static int inofd;
 static struct event_base *evbase_scan;
-static struct event inoev;
-static struct event exitev;
-static struct event cmdev;
+static struct event *inoev;
 static pthread_t tid_scan;
 static struct deferred_pl *playlists;
 static struct stacked_dir *dirstack;
+static struct commands_base *cmdbase;
+
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+struct deferred_file
+{
+  struct watch_info wi;
+  struct inotify_event ie;
+  char path[PATH_MAX];
+
+  struct deferred_file *next;
+};
+
+static struct deferred_file *filestack;
+static struct event *deferred_inoev;
+#endif
 
 /* Count of files scanned during a bulk scan */
 static int counter;
@@ -154,8 +140,8 @@ static int scanning;
  * when we get the IN_CREATE and then ignore the IN_ATTRIB for these files.
  */
 #define INCOMINGFILES_BUFFER_SIZE 50
-int incomingfiles_idx;
-uint32_t incomingfiles_buffer[INCOMINGFILES_BUFFER_SIZE];
+static int incomingfiles_idx;
+static uint32_t incomingfiles_buffer[INCOMINGFILES_BUFFER_SIZE];
 
 /* Forward */
 static void
@@ -164,49 +150,14 @@ static int
 inofd_event_set(void);
 static void
 inofd_event_unset(void);
-static int
-filescanner_initscan(struct filescanner_command *cmd);
-static int
-filescanner_fullrescan(struct filescanner_command *cmd);
+static enum command_state
+filescanner_initscan(void *arg, int *retval);
+static enum command_state
+filescanner_fullrescan(void *arg, int *retval);
 
 
-/* ---------------------------- COMMAND EXECUTION -------------------------- */
-
 static int
-send_command(struct filescanner_command *cmd)
-{
-  int ret;
-
-  if (!cmd->func)
-    {
-      DPRINTF(E_LOG, L_SCAN, "BUG: cmd->func is NULL!\n");
-      return -1;
-    }
-
-  ret = write(cmd_pipe[1], &cmd, sizeof(cmd));
-  if (ret != sizeof(cmd))
-    {
-      DPRINTF(E_LOG, L_SCAN, "Could not send command: %s\n", strerror(errno));
-      return -1;
-    }
-
-  return 0;
-}
-
-static int
-nonblock_command(struct filescanner_command *cmd)
-{
-  int ret;
-
-  ret = send_command(cmd);
-  if (ret < 0)
-    return -1;
-
-  return 0;
-}
-
-static int
-push_dir(struct stacked_dir **s, char *path)
+push_dir(struct stacked_dir **s, char *path, int parent_id)
 {
   struct stacked_dir *d;
 
@@ -224,28 +175,26 @@ push_dir(struct stacked_dir **s, char *path)
       return -1;
     }
 
+  d->parent_id = parent_id;
+
   d->next = *s;
   *s = d;
 
   return 0;
 }
 
-static char *
+static struct stacked_dir *
 pop_dir(struct stacked_dir **s)
 {
   struct stacked_dir *d;
-  char *ret;
 
   if (!*s)
     return NULL;
 
   d = *s;
   *s = d->next;
-  ret = d->path;
 
-  free(d);
-
-  return ret;
+  return d;
 }
 
 #ifdef HAVE_REGEX_H
@@ -325,6 +274,9 @@ file_type_get(const char *path) {
   if (!ext || (strlen(ext) == 1))
     return FILE_REGULAR;
 
+  if (file_type_ignore(ext))
+    return FILE_IGNORE;
+
   if ((strcasecmp(ext, ".m3u") == 0) || (strcasecmp(ext, ".pls") == 0))
     return FILE_PLAYLIST;
 
@@ -345,15 +297,11 @@ file_type_get(const char *path) {
   if (strcasecmp(ext, ".remote") == 0)
     return FILE_CTRL_REMOTE;
 
-#ifdef LASTFM
   if (strcasecmp(ext, ".lastfm") == 0)
     return FILE_CTRL_LASTFM;
-#endif
 
-#ifdef HAVE_SPOTIFY_H
   if (strcasecmp(ext, ".spotify") == 0)
     return FILE_CTRL_SPOTIFY;
-#endif
 
   if (strcasecmp(ext, ".init-rescan") == 0)
     return FILE_CTRL_INITSCAN;
@@ -367,9 +315,6 @@ file_type_get(const char *path) {
       return FILE_IGNORE;
     }
 
-  if (file_type_ignore(ext))
-    return FILE_IGNORE;
-
   if ((filename[0] == '_') || (filename[0] == '.'))
     return FILE_IGNORE;
 
@@ -642,7 +587,7 @@ fixup_tags(struct media_file_info *mfi)
 
 
 void
-filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct media_file_info *external_mfi)
+filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct media_file_info *external_mfi, int dir_id)
 {
   struct media_file_info *mfi;
   char *filename;
@@ -717,7 +662,7 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
       ret = scan_metadata_ffmpeg(path, mfi);
       if (ret < 0)
 	{
-	  DPRINTF(E_LOG, L_SCAN, "Playlist URL is unavailable for probe/metadata, assuming MP3 encoding\n");
+	  DPRINTF(E_LOG, L_SCAN, "Playlist URL '%s' is unavailable for probe/metadata, assuming MP3 encoding\n", path);
 	  mfi->type = strdup("mp3");
 	  mfi->codectype = strdup("mpeg");
 	  mfi->description = strdup("MPEG audio file");
@@ -739,13 +684,13 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
     }
   else
     {
-      DPRINTF(E_LOG, L_SCAN, "Unknown scan type for %s, this error should not occur\n", path);
+      DPRINTF(E_LOG, L_SCAN, "Unknown scan type for '%s', this error should not occur\n", path);
       ret = -1;
     }
 
   if (ret < 0)
     {
-      DPRINTF(E_INFO, L_SCAN, "Could not extract metadata for %s\n", path);
+      DPRINTF(E_INFO, L_SCAN, "Could not extract metadata for '%s'\n", path);
       goto out;
     }
 
@@ -774,6 +719,8 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
       mfi->virtual_path = strdup(virtual_path);
     }
 
+  mfi->directory_id = dir_id;
+
   if (mfi->id == 0)
     db_file_add(mfi);
   else
@@ -785,13 +732,13 @@ filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct
 }
 
 static void
-process_playlist(char *file, time_t mtime)
+process_playlist(char *file, time_t mtime, int dir_id)
 {
   enum file_type ft;
 
   ft = file_type_get(file);
   if (ft == FILE_PLAYLIST)
-    scan_playlist(file, mtime);
+    scan_playlist(file, mtime, dir_id);
 #ifdef ITUNES
   else if (ft == FILE_ITUNES)
     scan_itunes_itml(file);
@@ -800,7 +747,7 @@ process_playlist(char *file, time_t mtime)
 
 /* Thread: scan */
 static void
-defer_playlist(char *path, time_t mtime)
+defer_playlist(char *path, time_t mtime, int dir_id)
 {
   struct deferred_pl *pl;
 
@@ -824,6 +771,7 @@ defer_playlist(char *path, time_t mtime)
     }
 
   pl->mtime = mtime;
+  pl->directory_id = dir_id;
   pl->next = playlists;
   playlists = pl;
 
@@ -840,7 +788,7 @@ process_deferred_playlists(void)
     {
       playlists = pl->next;
 
-      process_playlist(pl->path, pl->mtime);
+      process_playlist(pl->path, pl->mtime, pl->directory_id);
 
       free(pl->path);
       free(pl);
@@ -852,16 +800,17 @@ process_deferred_playlists(void)
 
 /* Thread: scan */
 static void
-process_file(char *file, time_t mtime, off_t size, int type, int flags)
+process_file(char *file, time_t mtime, off_t size, int type, int flags, int dir_id)
 {
   int is_bulkscan;
+  int ret;
 
   is_bulkscan = (flags & F_SCAN_BULK);
 
   switch (file_type_get(file))
     {
       case FILE_REGULAR:
-	filescanner_process_media(file, mtime, size, type, NULL);
+	filescanner_process_media(file, mtime, size, type, NULL, dir_id);
 
 	cache_artwork_ping(file, mtime, !is_bulkscan);
 	// TODO [artworkcache] If entry in artwork cache exists for no artwork available, delete the entry if media file has embedded artwork
@@ -880,14 +829,14 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
       case FILE_PLAYLIST:
       case FILE_ITUNES:
 	if (flags & F_SCAN_BULK)
-	  defer_playlist(file, mtime);
+	  defer_playlist(file, mtime, dir_id);
 	else
-	  process_playlist(file, mtime);
+	  process_playlist(file, mtime, dir_id);
 	break;
 
       case FILE_SMARTPL:
 	DPRINTF(E_DBG, L_SCAN, "Smart playlist file: %s\n", file);
-	scan_smartpl(file, mtime);
+	scan_smartpl(file, mtime, dir_id);
 	break;
 
       case FILE_ARTWORK:
@@ -902,17 +851,21 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 	remote_pairing_read_pin(file);
 	break;
 
-#ifdef LASTFM
       case FILE_CTRL_LASTFM:
+#ifdef LASTFM
 	lastfm_login(file);
-	break;
+#else
+	DPRINTF(E_LOG, L_SCAN, "Detected LastFM file, but this version was built without LastFM support\n");
 #endif
+	break;
 
-#ifdef HAVE_SPOTIFY_H
       case FILE_CTRL_SPOTIFY:
+#ifdef HAVE_SPOTIFY_H
 	spotify_login(file);
-	break;
+#else
+	DPRINTF(E_LOG, L_SCAN, "Detected Spotify file, but this version was built without Spotify support\n");
 #endif
+	break;
 
       case FILE_CTRL_INITSCAN:
 	if (flags & F_SCAN_BULK)
@@ -920,7 +873,7 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 
 	DPRINTF(E_LOG, L_SCAN, "Startup rescan triggered, found init-rescan file: %s\n", file);
 
-	filescanner_initscan(NULL);
+	filescanner_initscan(NULL, &ret);
 	break;
 
       case FILE_CTRL_FULLSCAN:
@@ -929,7 +882,7 @@ process_file(char *file, time_t mtime, off_t size, int type, int flags)
 
 	DPRINTF(E_LOG, L_SCAN, "Full rescan triggered, found full-rescan file: %s\n", file);
 
-	filescanner_fullrescan(NULL);
+	filescanner_fullrescan(NULL, &ret);
 	break;
 
       default:
@@ -957,8 +910,22 @@ check_speciallib(char *path, const char *libtype)
 }
 
 /* Thread: scan */
+static int
+create_virtual_path(char *path, char *virtual_path, int virtual_path_len)
+{
+  int ret;
+  ret = snprintf(virtual_path, virtual_path_len, "/file:%s", path);
+  if ((ret < 0) || (ret >= virtual_path_len))
+  {
+    DPRINTF(E_LOG, L_SCAN, "Virtual path /file:%s, PATH_MAX exceeded\n", path);
+    return -1;
+  }
+
+  return 0;
+}
+
 static void
-process_directory(char *path, int flags)
+process_directory(char *path, int parent_id, int flags)
 {
   DIR *dirp;
   struct dirent buf;
@@ -967,10 +934,9 @@ process_directory(char *path, int flags)
   char *deref;
   struct stat sb;
   struct watch_info wi;
-#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-  struct kevent kev;
-#endif
   int type;
+  char virtual_path[PATH_MAX];
+  int dir_id;
   int ret;
 
   DPRINTF(E_DBG, L_SCAN, "Processing directory %s (flags = 0x%x)\n", path, flags);
@@ -983,6 +949,18 @@ process_directory(char *path, int flags)
       return;
     }
 
+  /* Add/update directories table */
+
+  ret = create_virtual_path(path, virtual_path, sizeof(virtual_path));
+  if (ret < 0)
+    return;
+
+  dir_id = db_directory_addorupdate(virtual_path, 0, parent_id);
+  if (dir_id <= 0)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Insert or update of directory failed '%s'\n", virtual_path);
+    }
+
   /* Check if compilation and/or podcast directory */
   type = 0;
   if (check_speciallib(path, "compilations"))
@@ -1059,15 +1037,15 @@ process_directory(char *path, int flags)
       if (S_ISREG(sb.st_mode))
 	{
 	  if (!(flags & F_SCAN_FAST))
-	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, flags);
+	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, flags, dir_id);
 	}
       else if (S_ISFIFO(sb.st_mode))
 	{
 	  if (!(flags & F_SCAN_FAST))
-	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, flags);
+	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, flags, dir_id);
 	}
       else if (S_ISDIR(sb.st_mode))
-	push_dir(&dirstack, entry);
+	push_dir(&dirstack, entry, dir_id);
       else
 	DPRINTF(E_LOG, L_SCAN, "Skipping %s, not a directory, symlink, pipe nor regular file\n", entry);
     }
@@ -1076,9 +1054,13 @@ process_directory(char *path, int flags)
 
   memset(&wi, 0, sizeof(struct watch_info));
 
+  // Add inotify watch (for FreeBSD we limit the flags so only dirs will be
+  // opened, otherwise we will be opening way too many files)
 #if defined(__linux__)
-  /* Add inotify watch */
   wi.wd = inotify_add_watch(inofd, path, IN_ATTRIB | IN_CREATE | IN_DELETE | IN_CLOSE_WRITE | IN_MOVE | IN_DELETE | IN_MOVE_SELF);
+#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+  wi.wd = inotify_add_watch(inofd, path, IN_CREATE | IN_DELETE | IN_MOVE);
+#endif
   if (wi.wd < 0)
     {
       DPRINTF(E_WARN, L_SCAN, "Could not create inotify watch for %s: %s\n", path, strerror(errno));
@@ -1093,53 +1075,66 @@ process_directory(char *path, int flags)
 
       db_watch_add(&wi);
     }
-
-#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-  memset(&kev, 0, sizeof(struct kevent));
-
-  wi.wd = open(path, O_RDONLY | O_NONBLOCK);
-  if (wi.wd < 0)
-    {
-      DPRINTF(E_WARN, L_SCAN, "Could not open directory %s for watching: %s\n", path, strerror(errno));
-
-      return;
-    }
-
-  /* Add kevent */
-  EV_SET(&kev, wi.wd, EVFILT_VNODE, EV_ADD | EV_CLEAR, NOTE_DELETE | NOTE_WRITE | NOTE_RENAME, 0, NULL);
-
-  ret = kevent(inofd, &kev, 1, NULL, 0, NULL);
-  if (ret < 0)
-    {
-      DPRINTF(E_WARN, L_SCAN, "Could not add kevent for %s: %s\n", path, strerror(errno));
-
-      close(wi.wd);
-      return;
-    }
-
-  wi.cookie = 0;
-  wi.path = path;
-
-  db_watch_add(&wi);
-#endif
 }
 
 /* Thread: scan */
-static void
-process_directories(char *root, int flags)
+static int
+process_parent_directories(char *path)
 {
-  char *path;
+  char *ptr;
+  int dir_id;
+  char buf[PATH_MAX];
+  char virtual_path[PATH_MAX];
+  int ret;
 
-  process_directory(root, flags);
+  dir_id = DIR_FILE;
+
+  ptr = path + 1;
+  while (ptr && (ptr = strchr(ptr, '/')))
+    {
+      if (strlen(ptr) <= 1)
+	{
+	  // Do not process trailing '/'
+	  break;
+	}
+
+      strncpy(buf, path, (ptr - path));
+      buf[(ptr - path)] = '\0';
+
+      ret = create_virtual_path(buf, virtual_path, sizeof(virtual_path));
+      if (ret < 0)
+	return 0;
+
+      dir_id = db_directory_addorupdate(virtual_path, 0, dir_id);
+      if (dir_id <= 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Insert or update of directory failed '%s'\n", virtual_path);
+
+	  return 0;
+	}
+
+      ptr++;
+    }
+
+  return dir_id;
+}
+
+static void
+process_directories(char *root, int parent_id, int flags)
+{
+  struct stacked_dir *dir;
+
+  process_directory(root, parent_id, flags);
 
   if (scan_exit)
     return;
 
-  while ((path = pop_dir(&dirstack)))
+  while ((dir = pop_dir(&dirstack)))
     {
-      process_directory(path, flags);
+      process_directory(dir->path, dir->parent_id, flags);
 
-      free(path);
+      free(dir->path);
+      free(dir);
 
       if (scan_exit)
 	return;
@@ -1157,6 +1152,7 @@ bulk_scan(int flags)
   char *deref;
   time_t start;
   time_t end;
+  int parent_id;
   int i;
 
   // Set global flag to avoid queued scan requests
@@ -1174,6 +1170,8 @@ bulk_scan(int flags)
     {
       path = cfg_getnstr(lib, "directories", i);
 
+      parent_id = process_parent_directories(path);
+
       deref = m_realpath(path);
       if (!deref)
 	{
@@ -1182,16 +1180,19 @@ bulk_scan(int flags)
 	  /* Assume dir is mistakenly not mounted, so just disable everything and update timestamps */
 	  db_file_disable_bymatch(path, "", 0);
 	  db_pl_disable_bymatch(path, "", 0);
+	  db_directory_disable_bymatch(path, "", 0);
 
 	  db_file_ping_bymatch(path, 1);
 	  db_pl_ping_bymatch(path, 1);
+	  db_directory_ping_bymatch(path);
 
 	  continue;
 	}
 
       counter = 0;
       db_transaction_begin();
-      process_directories(deref, flags);
+
+      process_directories(deref, parent_id, flags);
       db_transaction_end();
 
       free(deref);
@@ -1302,7 +1303,7 @@ filescanner(void *arg)
 #endif
 
       /* Enable inotify */
-      event_add(&inoev, NULL);
+      event_add(inoev, NULL);
 
       event_base_dispatch(evbase_scan);
     }
@@ -1315,8 +1316,28 @@ filescanner(void *arg)
   pthread_exit(NULL);
 }
 
+static int
+get_parent_dir_id(const char *path)
+{
+  char *pathcopy;
+  char *parent_dir;
+  char virtual_path[PATH_MAX];
+  int parent_id;
+  int ret;
+
+  pathcopy = strdup(path);
+  parent_dir = dirname(pathcopy);
+  ret = create_virtual_path(parent_dir, virtual_path, sizeof(virtual_path));
+  if (ret == 0)
+    parent_id = db_directory_id_byvirtualpath(virtual_path);
+  else
+    parent_id = 0;
+
+  free(pathcopy);
+
+  return parent_id;
+}
 
-#if defined(__linux__)
 static int
 watches_clear(uint32_t wd, char *path)
 {
@@ -1356,13 +1377,15 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
   char *s;
   int flags = 0;
   int ret;
+  int parent_id;
 
-  DPRINTF(E_SPAM, L_SCAN, "Directory event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
+  DPRINTF(E_DBG, L_SCAN, "Directory event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
 
   if (ie->mask & IN_UNMOUNT)
     {
       db_file_disable_bymatch(path, "", 0);
       db_pl_disable_bymatch(path, "", 0);
+      db_directory_disable_bymatch(path, "", 0);
     }
 
   if (ie->mask & IN_MOVE_SELF)
@@ -1417,6 +1440,7 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
       db_watch_mark_bymatch(path, path, ie->cookie);
       db_file_disable_bymatch(path, path, ie->cookie);
       db_pl_disable_bymatch(path, path, ie->cookie);
+      db_directory_disable_bymatch(path, path, ie->cookie);
     }
 
   if (ie->mask & IN_MOVED_TO)
@@ -1426,6 +1450,7 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 	  db_watch_move_bycookie(ie->cookie, path);
 	  db_file_enable_bycookie(ie->cookie, path);
 	  db_pl_enable_bycookie(ie->cookie, path);
+	  db_directory_enable_bycookie(ie->cookie, path);
 
 	  /* We'll rescan the directory tree to update playlists */
 	  flags |= F_SCAN_MOVED;
@@ -1459,6 +1484,7 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 
 	  db_file_disable_bymatch(path, "", 0);
 	  db_pl_disable_bymatch(path, "", 0);
+	  db_directory_disable_bymatch(path, "", 0);
 	}
       else if (ret < 0)
 	{
@@ -1474,7 +1500,8 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 
   if (ie->mask & IN_CREATE)
     {
-      process_directories(path, flags);
+      parent_id = get_parent_dir_id(path);
+      process_directories(path, parent_id, flags);
 
       if (dirstack)
 	DPRINTF(E_LOG, L_SCAN, "WARNING: unhandled leftover directories\n");
@@ -1489,11 +1516,15 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
   uint32_t path_hash;
   char *deref = NULL;
   char *file = path;
+  char *dir;
+  char dir_vpath[PATH_MAX];
   int type;
   int i;
+  int dir_id;
+  char *ptr;
   int ret;
 
-  DPRINTF(E_SPAM, L_SCAN, "File event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
+  DPRINTF(E_DBG, L_SCAN, "File event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
 
   path_hash = djb_hash(path, strlen(path));
 
@@ -1550,7 +1581,28 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 
       ret = db_file_enable_bycookie(ie->cookie, path);
 
-      if (ret <= 0)
+      if (ret > 0)
+	{
+	  // If file was successfully enabled, update the directory id
+	  dir = strdup(path);
+	  ptr = strrchr(dir, '/');
+	  dir[(ptr - dir)] = '\0';
+
+	  ret = create_virtual_path(dir, dir_vpath, sizeof(dir_vpath));
+	  if (ret >= 0)
+	    {
+	      dir_id = db_directory_id_byvirtualpath(dir_vpath);
+	      if (dir_id > 0)
+		{
+		  ret = db_file_update_directoryid(path, dir_id);
+		  if (ret < 0)
+		    DPRINTF(E_LOG, L_SCAN, "Error updating directory id for file: %s\n", path);
+		}
+	    }
+
+	  free(dir);
+	}
+      else
 	{
 	  /* It's not a known media file, so it's either a new file
 	   * or a playlist, known or not.
@@ -1645,30 +1697,87 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
       if (check_speciallib(path, "audiobooks"))
 	type |= F_SCAN_TYPE_AUDIOBOOK;
 
+      dir_id = get_parent_dir_id(file);
+
       if (S_ISREG(sb.st_mode))
-	process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, 0);
+	{
+	  process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, 0, dir_id);
+	}
       else if (S_ISFIFO(sb.st_mode))
-	process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, 0);
+	process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, 0, dir_id);
 
       if (deref)
 	free(deref);
     }
 }
 
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+/* Since FreeBSD doesn't really have inotify we only get a IN_CREATE. That is
+ * a bit too soon to start scanning the file, so we defer it for 10 seconds.
+ */
+static void
+inotify_deferred_cb(int fd, short what, void *arg)
+{
+  struct deferred_file *f;
+  struct deferred_file *next;
+
+  for (f = filestack; f; f = next)
+    {
+      next = f->next;
+
+      DPRINTF(E_DBG, L_SCAN, "Processing deferred file %s\n", f->path);
+      process_inotify_file(&f->wi, f->path, &f->ie);
+      free(f->wi.path);
+      free(f);
+    }
+
+  filestack = NULL;
+}
+
+static void
+process_inotify_file_defer(struct watch_info *wi, char *path, struct inotify_event *ie)
+{
+  struct deferred_file *f;
+  struct timeval tv = { 10, 0 };
+
+  if (!(ie->mask & IN_CREATE))
+    {
+      process_inotify_file(wi, path, ie);
+      return;
+    }
+
+  DPRINTF(E_INFO, L_SCAN, "Deferring scan of newly created file %s\n", path);
+
+  ie->mask = IN_CLOSE_WRITE;
+  f = calloc(1, sizeof(struct deferred_file));
+  f->wi = *wi;
+  f->wi.path = strdup(wi->path);
+  f->ie = *ie;
+  strcpy(f->path, path);
+
+  f->next = filestack;
+  filestack = f;
+
+  event_add(deferred_inoev, &tv);
+}
+#endif
+
+
 /* Thread: scan */
 static void
 inotify_cb(int fd, short event, void *arg)
 {
-  struct inotify_event *buf;
   struct inotify_event *ie;
   struct watch_info wi;
+  uint8_t *buf;
+  uint8_t *ptr;
   char path[PATH_MAX];
-  int qsize;
+  int size;
   int namelen;
   int ret;
 
-  /* Determine the size of the inotify queue */
-  ret = ioctl(fd, FIONREAD, &qsize);
+  /* Determine the amount of bytes to read from inotify */
+  ret = ioctl(fd, FIONREAD, &size);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_SCAN, "Could not determine inotify queue size: %s\n", strerror(errno));
@@ -1676,29 +1785,27 @@ inotify_cb(int fd, short event, void *arg)
       return;
     }
 
-  buf = (struct inotify_event *)malloc(qsize);
+  buf = malloc(size);
   if (!buf)
     {
-      DPRINTF(E_LOG, L_SCAN, "Could not allocate %d bytes for inotify events\n", qsize);
+      DPRINTF(E_LOG, L_SCAN, "Could not allocate %d bytes for inotify events\n", size);
 
       return;
     }
 
-  ret = read(fd, buf, qsize);
-  if (ret < 0)
+  ret = read(fd, buf, size);
+  if (ret < 0 || ret != size)
     {
-      DPRINTF(E_LOG, L_SCAN, "inotify read failed: %s\n", strerror(errno));
+      DPRINTF(E_LOG, L_SCAN, "inotify read failed: %s (ret was %d, size %d)\n", strerror(errno), ret, size);
 
       free(buf);
       return;
     }
 
-  /* ioctl(FIONREAD) returns the number of bytes, now we need the number of elements */
-  qsize /= sizeof(struct inotify_event);
-
-  /* Loop through all the events we got */
-  for (ie = buf; (ie - buf) < qsize; ie += (1 + (ie->len / sizeof(struct inotify_event))))
+  for (ptr = buf; ptr < buf + size; ptr += ie->len + sizeof(struct inotify_event))
     {
+      ie = (struct inotify_event *)ptr;
+
       memset(&wi, 0, sizeof(struct watch_info));
 
       /* ie[0] contains the inotify event information
@@ -1756,205 +1863,23 @@ inotify_cb(int fd, short event, void *arg)
       if ((ie->mask & IN_ISDIR) || (ie->len == 0))
 	process_inotify_dir(&wi, path, ie);
       else
+#if defined(__linux__)
 	process_inotify_file(&wi, path, ie);
-
+#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+	process_inotify_file_defer(&wi, path, ie);
+#endif
       free(wi.path);
     }
 
   free(buf);
 
-  event_add(&inoev, NULL);
+  event_add(inoev, NULL);
 }
-#endif /* __linux__ */
-
-
-#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-/* Thread: scan */
-static void
-kqueue_cb(int fd, short event, void *arg)
-{
-  struct kevent kev;
-  struct timespec ts;
-  struct watch_info wi;
-  struct watch_enum we;
-  struct stacked_dir *rescan;
-  struct stacked_dir *d;
-  struct stacked_dir *dprev;
-  char *path;
-  uint32_t wd;
-  int d_len;
-  int w_len;
-  int need_rescan;
-  int ret;
-
-  ts.tv_sec = 0;
-  ts.tv_nsec = 0;
-
-  we.cookie = 0;
-
-  rescan = NULL;
-
-  DPRINTF(E_DBG, L_SCAN, "Library changed!\n");
-
-  /* We can only monitor directories with kqueue; to monitor files, we'd need
-   * to have an open fd on every file in the library, which is totally insane.
-   * Unfortunately, that means we only know when directories get renamed,
-   * deleted or changed. We don't get directory/file names when directories/files
-   * are created/deleted/renamed in the directory, so we have to rescan.
-   */
-  while (kevent(fd, NULL, 0, &kev, 1, &ts) > 0)
-    {
-      /* This should not happen, and if it does, we'll end up in
-       * an infinite loop.
-       */
-      if (kev.filter != EVFILT_VNODE)
-	continue;
-
-      wi.wd = kev.ident;
-
-      ret = db_watch_get_bywd(&wi);
-      if (ret < 0)
-	{
-	  DPRINTF(E_LOG, L_SCAN, "Found no matching watch for kevent, killing this event\n");
-
-	  close(kev.ident);
-	  continue;
-	}
-
-      /* Whatever the type of event that happened, disable matching watches and
-       * files before we trigger an eventual rescan.
-       */
-      we.match = wi.path;
-
-      ret = db_watch_enum_start(&we);
-      if (ret < 0)
-	{
-	  free(wi.path);
-	  continue;
-	}
-
-      while ((db_watch_enum_fetchwd(&we, &wd) == 0) && (wd))
-	{
-	  close(wd);
-	}
-
-      db_watch_enum_end(&we);
-
-      db_watch_delete_bymatch(wi.path);
-
-      close(wi.wd);
-      db_watch_delete_bywd(wi.wd);
-
-      /* Disable files */
-      db_file_disable_bymatch(wi.path, "", 0);
-      db_pl_disable_bymatch(wi.path, "", 0);
-
-      if (kev.flags & EV_ERROR)
-	{
-	  DPRINTF(E_LOG, L_SCAN, "kevent reports EV_ERROR (%s): %s\n", wi.path, strerror(kev.data));
-
-	  ret = access(wi.path, F_OK);
-	  if (ret != 0)
-	    {
-	      free(wi.path);
-	      continue;
-	    }
-
-	  /* The directory still exists, so try to add it back to the library */
-	  kev.fflags |= NOTE_WRITE;
-	}
-
-      /* No further action on NOTE_DELETE & NOTE_RENAME; NOTE_WRITE on the
-       * parent directory will trigger a rescan in both cases and the
-       * renamed directory will be picked up then.
-       */
-
-      if (kev.fflags & NOTE_WRITE)
-	{
-          DPRINTF(E_DBG, L_SCAN, "Got NOTE_WRITE (%s)\n", wi.path);
-
-	  need_rescan = 1;
-	  w_len = strlen(wi.path);
-
-	  /* Abusing stacked_dir a little bit here */
-	  dprev = NULL;
-	  d = rescan;
-	  while (d)
-	    {
-	      d_len = strlen(d->path);
-
-	      if (d_len > w_len)
-		{
-		  /* Stacked dir child of watch dir? */
-		  if ((d->path[w_len] == '/') && (strncmp(d->path, wi.path, w_len) == 0))
-		    {
-		      DPRINTF(E_DBG, L_SCAN, "Watched directory is a parent\n");
-
-		      if (dprev)
-			dprev->next = d->next;
-		      else
-			rescan = d->next;
-
-		      free(d->path);
-		      free(d);
-
-		      if (dprev)
-			d = dprev->next;
-		      else
-			d = rescan;
-
-		      continue;
-		    }
-		}
-	      else if (w_len > d_len)
-		{
-		  /* Watch dir child of stacked dir? */
-		  if ((wi.path[d_len] == '/') && (strncmp(wi.path, d->path, d_len) == 0))
-		    {
-		      DPRINTF(E_DBG, L_SCAN, "Watched directory is a child\n");
-
-		      need_rescan = 0;
-		      break;
-		    }
-		}
-	      else if (strcmp(wi.path, d->path) == 0)
-		{
-		  DPRINTF(E_DBG, L_SCAN, "Watched directory already listed\n");
-
-		  need_rescan = 0;
-		  break;
-		}
-
-	      dprev = d;
-	      d = d->next;
-	    }
-
-	  if (need_rescan)
-	    push_dir(&rescan, wi.path);
-	}
-
-      free(wi.path);
-    }
-
-  while ((path = pop_dir(&rescan)))
-    {
-      process_directories(path, 0);
-
-      free(path);
-
-      if (rescan)
-	DPRINTF(E_LOG, L_SCAN, "WARNING: unhandled leftover directories\n");
-    }
-
-  event_add(&inoev, NULL);
-}
-#endif /* __FreeBSD__ || __FreeBSD_kernel__ */
 
 /* Thread: main & scan */
 static int
 inofd_event_set(void)
 {
-#if defined(__linux__)
   inofd = inotify_init1(IN_CLOEXEC);
   if (inofd < 0)
     {
@@ -1963,23 +1888,18 @@ inofd_event_set(void)
       return -1;
     }
 
-  event_set(&inoev, inofd, EV_READ, inotify_cb, NULL);
+  inoev = event_new(evbase_scan, inofd, EV_READ, inotify_cb, NULL);
 
-#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-
-  inofd = kqueue();
-  if (inofd < 0)
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+  deferred_inoev = evtimer_new(evbase_scan, inotify_deferred_cb, NULL);
+  if (!deferred_inoev)
     {
-      DPRINTF(E_FATAL, L_SCAN, "Could not create kqueue: %s\n", strerror(errno));
+      DPRINTF(E_LOG, L_SCAN, "Could not create deferred inotify event\n");
 
       return -1;
     }
-
-  event_set(&inoev, inofd, EV_READ, kqueue_cb, NULL);
 #endif
 
-  event_base_set(evbase_scan, &inoev);
-
   return 0;
 }
 
@@ -1987,54 +1907,17 @@ inofd_event_set(void)
 static void
 inofd_event_unset(void)
 {
-  event_del(&inoev);
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+  event_free(deferred_inoev);
+#endif
+  event_free(inoev);
   close(inofd);
 }
 
 /* Thread: scan */
-static void
-exit_cb(int fd, short event, void *arg)
-{
-  event_base_loopbreak(evbase_scan);
 
-  scan_exit = 1;
-}
-
-static void
-command_cb(int fd, short what, void *arg)
-{
-  struct filescanner_command *cmd;
-  int ret;
-
-  ret = read(cmd_pipe[0], &cmd, sizeof(cmd));
-  if (ret != sizeof(cmd))
-    {
-      DPRINTF(E_LOG, L_SCAN, "Could not read command! (read %d): %s\n", ret, (ret < 0) ? strerror(errno) : "-no error-");
-      goto readd;
-    }
-
-  if (cmd->nonblock)
-    {
-      cmd->func(cmd);
-
-      free(cmd);
-      goto readd;
-    }
-
-  pthread_mutex_lock(&cmd->lck);
-
-  ret = cmd->func(cmd);
-  cmd->ret = ret;
-
-  pthread_cond_signal(&cmd->cond);
-  pthread_mutex_unlock(&cmd->lck);
-
- readd:
-  event_add(&cmdev, NULL);
-}
-
-static int
-filescanner_initscan(struct filescanner_command *cmd)
+static enum command_state
+filescanner_initscan(void *arg, int *retval)
 {
   DPRINTF(E_LOG, L_SCAN, "Startup rescan triggered\n");
 
@@ -2044,11 +1927,12 @@ filescanner_initscan(struct filescanner_command *cmd)
   inofd_event_set();
   bulk_scan(F_SCAN_BULK | F_SCAN_RESCAN);
 
-  return 0;
+  *retval = 0;
+  return COMMAND_END;
 }
 
-static int
-filescanner_fullrescan(struct filescanner_command *cmd)
+static enum command_state
+filescanner_fullrescan(void *arg, int *retval)
 {
   DPRINTF(E_LOG, L_SCAN, "Full rescan triggered\n");
 
@@ -2060,62 +1944,32 @@ filescanner_fullrescan(struct filescanner_command *cmd)
   inofd_event_set();
   bulk_scan(F_SCAN_BULK);
 
-  return 0;
+  *retval = 0;
+  return COMMAND_END;
 }
 
 void
 filescanner_trigger_initscan(void)
 {
-  struct filescanner_command *cmd;
-
   if (scanning)
     {
       DPRINTF(E_INFO, L_SCAN, "Scan already running, ignoring request to trigger a new init scan\n");
       return;
     }
 
-
-  cmd = (struct filescanner_command *)malloc(sizeof(struct filescanner_command));
-  if (!cmd)
-    {
-      DPRINTF(E_LOG, L_SCAN, "Could not allocate cache_command\n");
-      return;
-    }
-
-  memset(cmd, 0, sizeof(struct filescanner_command));
-
-  cmd->nonblock = 1;
-
-  cmd->func = filescanner_initscan;
-
-  nonblock_command(cmd);
+ commands_exec_async(cmdbase, filescanner_initscan, NULL);
 }
 
 void
 filescanner_trigger_fullrescan(void)
 {
-  struct filescanner_command *cmd;
-
   if (scanning)
     {
       DPRINTF(E_INFO, L_SCAN, "Scan already running, ignoring request to trigger a new init scan\n");
       return;
     }
 
-  cmd = (struct filescanner_command *)malloc(sizeof(struct filescanner_command));
-  if (!cmd)
-    {
-      DPRINTF(E_LOG, L_SCAN, "Could not allocate cache_command\n");
-      return;
-    }
-
-  memset(cmd, 0, sizeof(struct filescanner_command));
-
-  cmd->nonblock = 1;
-
-  cmd->func = filescanner_fullrescan;
-
-  nonblock_command(cmd);
+  commands_exec_async(cmdbase, filescanner_fullrescan, NULL);
 }
 
 /*
@@ -2145,56 +1999,13 @@ filescanner_init(void)
       return -1;
     }
 
-#ifdef USE_EVENTFD
-  exit_efd = eventfd(0, EFD_CLOEXEC);
-  if (exit_efd < 0)
-    {
-      DPRINTF(E_FATAL, L_SCAN, "Could not create eventfd: %s\n", strerror(errno));
-
-      goto pipe_fail;
-    }
-#else
-# if defined(__linux__)
-  ret = pipe2(exit_pipe, O_CLOEXEC);
-# else
-  ret = pipe(exit_pipe);
-# endif
-  if (ret < 0)
-    {
-      DPRINTF(E_FATAL, L_SCAN, "Could not create pipe: %s\n", strerror(errno));
-
-      goto pipe_fail;
-    }
-#endif /* USE_EVENTFD */
-
   ret = inofd_event_set();
   if (ret < 0)
     {
       goto ino_fail;
     }
 
-#ifdef USE_EVENTFD
-  event_set(&exitev, exit_efd, EV_READ, exit_cb, NULL);
-#else
-  event_set(&exitev, exit_pipe[0], EV_READ, exit_cb, NULL);
-#endif
-  event_base_set(evbase_scan, &exitev);
-  event_add(&exitev, NULL);
-
-# if defined(__linux__)
-  ret = pipe2(cmd_pipe, O_CLOEXEC);
-# else
-  ret = pipe(cmd_pipe);
-# endif
-  if (ret < 0)
-    {
-      DPRINTF(E_LOG, L_SCAN, "Could not create command pipe: %s\n", strerror(errno));
-      goto cmd_fail;
-    }
-
-  event_set(&cmdev, cmd_pipe[0], EV_READ, command_cb, NULL);
-  event_base_set(evbase_scan, &cmdev);
-  event_add(&cmdev, NULL);
+  cmdbase = commands_base_new(evbase_scan, NULL);
 
   ret = pthread_create(&tid_scan, NULL, filescanner, NULL);
   if (ret != 0)
@@ -2204,21 +2015,18 @@ filescanner_init(void)
       goto thread_fail;
     }
 
+#if defined(HAVE_PTHREAD_SETNAME_NP)
+  pthread_setname_np(tid_scan, "filescanner");
+#elif defined(HAVE_PTHREAD_SET_NAME_NP)
+  pthread_set_name_np(tid_scan, "filescanner");
+#endif
+
   return 0;
 
  thread_fail:
- cmd_fail:
-  close(cmd_pipe[0]);
-  close(cmd_pipe[1]);
+  commands_base_free(cmdbase);
   close(inofd);
  ino_fail:
-#ifdef USE_EVENTFD
-  close(exit_efd);
-#else
-  close(exit_pipe[0]);
-  close(exit_pipe[1]);
-#endif
- pipe_fail:
   event_base_free(evbase_scan);
 
   return -1;
@@ -2230,27 +2038,8 @@ filescanner_deinit(void)
 {
   int ret;
 
-#ifdef USE_EVENTFD
-  ret = eventfd_write(exit_efd, 1);
-  if (ret < 0)
-    {
-      DPRINTF(E_FATAL, L_SCAN, "Could not send exit event: %s\n", strerror(errno));
-
-      return;
-    }
-#else
-  int dummy = 42;
-
-  ret = write(exit_pipe[1], &dummy, sizeof(dummy));
-  if (ret != sizeof(dummy))
-    {
-      DPRINTF(E_FATAL, L_SCAN, "Could not write to exit fd: %s\n", strerror(errno));
-
-      return;
-    }
-#endif
-
   scan_exit = 1;
+  commands_base_destroy(cmdbase);
 
   ret = pthread_join(tid_scan, NULL);
   if (ret != 0)
@@ -2262,13 +2051,5 @@ filescanner_deinit(void)
 
   inofd_event_unset();
 
-#ifdef USE_EVENTFD
-  close(exit_efd);
-#else
-  close(exit_pipe[0]);
-  close(exit_pipe[1]);
-#endif
-  close(cmd_pipe[0]);
-  close(cmd_pipe[1]);
   event_base_free(evbase_scan);
 }
