@@ -74,7 +74,7 @@ extinf_get(char *string, struct media_file_info *mfi, int *extinf)
 }
 
 void
-scan_playlist(char *file, time_t mtime)
+scan_playlist(char *file, time_t mtime, int dir_id)
 {
   FILE *fp;
   struct media_file_info mfi;
@@ -172,6 +172,8 @@ scan_playlist(char *file, time_t mtime)
 	*ptr = '\0';
       pli->virtual_path = strdup(virtual_path);
 
+      pli->directory_id = dir_id;
+
       ret = db_pl_add(pli, &pl_id);
       if (ret < 0)
 	{
@@ -236,7 +238,7 @@ scan_playlist(char *file, time_t mtime)
 	  if (extinf)
 	    DPRINTF(E_INFO, L_SCAN, "Playlist has EXTINF metadata, artist is '%s', title is '%s'\n", mfi.artist, mfi.title);
 
-	  filescanner_process_media(filename, mtime, 0, F_SCAN_TYPE_URL, &mfi);
+	  filescanner_process_media(filename, mtime, 0, F_SCAN_TYPE_URL, &mfi, DIR_HTTP);
 	}
       /* Regular file, should already be in library */
       else
