@@ -28,6 +28,9 @@
 #include <limits.h>
 #include <errno.h>
 #include <pthread.h>
+#ifdef HAVE_PTHREAD_NP_H
+# include <pthread_np.h>
+#endif
 #include <time.h>
 #include <sys/param.h>
 #include <sys/queue.h>
@@ -40,18 +43,27 @@
 # define USE_EVENTFD
 # include <sys/eventfd.h>
 #endif
-
+#include <event2/event.h>
+#ifdef HAVE_LIBEVENT2_OLD
+# include <event2/bufferevent.h>
+# include <event2/bufferevent_struct.h>
+#endif
 #include <zlib.h>
 
 #include "logger.h"
 #include "db.h"
 #include "conffile.h"
 #include "misc.h"
+#include "worker.h"
 #include "httpd.h"
 #include "httpd_rsp.h"
 #include "httpd_daap.h"
 #include "httpd_dacp.h"
+#include "httpd_streaming.h"
 #include "transcode.h"
+#ifdef LASTFM
+# include "lastfm.h"
+#endif
 
 /*
  * HTTP client quirks by User-Agent, from mt-daapd
@@ -74,6 +86,11 @@
 
 #define STREAM_CHUNK_SIZE (64 * 1024)
 #define WEBFACE_ROOT   DATADIR "/webface/"
+#define ERR_PAGE "<html>\n<head>\n" \
+  "<title>%d %s</title>\n" \
+  "</head>\n<body>\n" \
+  "<h1>%s</h1>\n" \
+  "</body>\n</html>\n"
 
 struct content_type_map {
   char *ext;
@@ -84,7 +101,7 @@ struct stream_ctx {
   struct evhttp_request *req;
   uint8_t *buf;
   struct evbuffer *evbuf;
-  struct event ev;
+  struct event *ev;
   int id;
   int fd;
   off_t size;
@@ -118,14 +135,17 @@ static int exit_efd;
 static int exit_pipe[2];
 #endif
 static int httpd_exit;
-static struct event exitev;
+static struct event *exitev;
 static struct evhttp *evhttpd;
 static pthread_t tid_httpd;
 
+static char *allow_origin;
+
 #ifdef HAVE_LIBEVENT2_OLD
 struct stream_ctx *g_st;
 #endif
 
+
 static void
 stream_end(struct stream_ctx *st, int failed)
 {
@@ -140,6 +160,7 @@ stream_end(struct stream_ctx *st, int failed)
     evhttp_send_reply_end(st->req);
 
   evbuffer_free(st->evbuf);
+  event_free(st->ev);
 
   if (st->xcode)
     transcode_cleanup(st->xcode);
@@ -157,15 +178,38 @@ stream_end(struct stream_ctx *st, int failed)
   free(st);
 }
 
+/* Callback from the worker thread (async operation as it may block) */
 static void
-stream_up_playcount(struct stream_ctx *st)
+playcount_inc_cb(void *arg)
+{
+  int *id = arg;
+
+  db_file_inc_playcount(*id);
+}
+
+#ifdef LASTFM
+/* Callback from the worker thread (async operation as it may block) */
+static void
+scrobble_cb(void *arg)
+{
+  int *id = arg;
+
+  lastfm_scrobble(*id);
+}
+#endif
+
+static void
+stream_end_register(struct stream_ctx *st)
 {
   if (!st->marked
       && (st->stream_size > ((st->size * 50) / 100))
       && (st->offset > ((st->size * 80) / 100)))
     {
       st->marked = 1;
-      db_file_inc_playcount(st->id);
+      worker_execute(playcount_inc_cb, &st->id, sizeof(int), 0);
+#ifdef LASTFM
+      worker_execute(scrobble_cb, &st->id, sizeof(int), 1);
+#endif
     }
 }
 
@@ -179,7 +223,7 @@ stream_chunk_resched_cb(struct evhttp_connection *evcon, void *arg)
   st = (struct stream_ctx *)arg;
 
   evutil_timerclear(&tv);
-  ret = event_add(&st->ev, &tv);
+  ret = event_add(st->ev, &tv);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not re-add one-shot event for streaming\n");
@@ -208,7 +252,7 @@ stream_chunk_xcode_cb(int fd, short event, void *arg)
 
   st = (struct stream_ctx *)arg;
 
-  xcoded = transcode(st->xcode, st->evbuf, STREAM_CHUNK_SIZE, &dummy);
+  xcoded = transcode(st->evbuf, STREAM_CHUNK_SIZE, st->xcode, &dummy);
   if (xcoded <= 0)
     {
       if (xcoded == 0)
@@ -259,13 +303,13 @@ stream_chunk_xcode_cb(int fd, short event, void *arg)
 
   st->offset += ret;
 
-  stream_up_playcount(st);
+  stream_end_register(st);
 
   return;
 
  consume: /* reschedule immediately - consume up to start_offset */
   evutil_timerclear(&tv);
-  ret = event_add(&st->ev, &tv);
+  ret = event_add(st->ev, &tv);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not re-add one-shot event for streaming (xcode)\n");
@@ -325,7 +369,7 @@ stream_chunk_raw_cb(int fd, short event, void *arg)
 
   st->offset += ret;
 
-  stream_up_playcount(st);
+  stream_end_register(st);
 }
 
 static void
@@ -338,7 +382,7 @@ stream_fail_cb(struct evhttp_connection *evcon, void *arg)
   DPRINTF(E_WARN, L_HTTPD, "Connection failed; stopping streaming of file ID %d\n", st->id);
 
   /* Stop streaming */
-  event_del(&st->ev);
+  event_del(st->ev);
 
   stream_end(st, 1);
 }
@@ -447,8 +491,8 @@ httpd_stream_file(struct evhttp_request *req, int id)
 
       stream_cb = stream_chunk_xcode_cb;
 
-      ret = transcode_setup(&st->xcode, mfi, &st->size, 1);
-      if (ret < 0)
+      st->xcode = transcode_setup(mfi, XCODE_PCM16_HEADER, &st->size);
+      if (!st->xcode)
 	{
 	  DPRINTF(E_WARN, L_HTTPD, "Transcoding setup failed, aborting streaming\n");
 
@@ -562,11 +606,9 @@ httpd_stream_file(struct evhttp_request *req, int id)
       goto out_cleanup;
     }
 
+  st->ev = event_new(evbase_httpd, -1, EV_TIMEOUT, stream_cb, st);
   evutil_timerclear(&tv);
-  event_set(&st->ev, -1, EV_TIMEOUT, stream_cb, st);
-  event_base_set(evbase_httpd, &st->ev);
-  ret = event_add(&st->ev, &tv);
-  if (ret < 0)
+  if (!st->ev || (event_add(st->ev, &tv) < 0))
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not add one-shot event for streaming\n");
 
@@ -658,130 +700,141 @@ httpd_stream_file(struct evhttp_request *req, int id)
   free_mfi(mfi, 0);
 }
 
-/* Thread: httpd */
-void
-httpd_send_reply(struct evhttp_request *req, int code, const char *reason, struct evbuffer *evbuf)
+struct evbuffer *
+httpd_gzip_deflate(struct evbuffer *in)
 {
-  unsigned char outbuf[128 * 1024];
+  struct evbuffer *out;
+  struct evbuffer_iovec iovec[1];
   z_stream strm;
-  struct evbuffer *gzbuf;
-  struct evkeyvalq *headers;
-  const char *param;
-  int flush;
-  int zret;
   int ret;
 
-  if (!req)
-    return;
-
-  if (!evbuf || (EVBUFFER_LENGTH(evbuf) == 0))
-    {
-      DPRINTF(E_DBG, L_HTTPD, "Not gzipping body-less reply\n");
-
-      goto no_gzip;
-    }
-
-  headers = evhttp_request_get_input_headers(req);
-
-  param = evhttp_find_header(headers, "Accept-Encoding");
-  if (!param)
-    {
-      DPRINTF(E_DBG, L_HTTPD, "Not gzipping; no Accept-Encoding header\n");
-
-      goto no_gzip;
-    }
-  else if (!strstr(param, "gzip") && !strstr(param, "*"))
-    {
-      DPRINTF(E_DBG, L_HTTPD, "Not gzipping; gzip not in Accept-Encoding (%s)\n", param);
-
-      goto no_gzip;
-    }
-
-  gzbuf = evbuffer_new();
-  if (!gzbuf)
-    {
-      DPRINTF(E_LOG, L_HTTPD, "Could not allocate evbuffer for gzipped reply\n");
-
-      goto no_gzip;
-    }
-
   strm.zalloc = Z_NULL;
   strm.zfree = Z_NULL;
   strm.opaque = Z_NULL;
 
-  /* Set up a gzip stream (the "+ 16" in 15 + 16), instead of a zlib stream (default) */
-  zret = deflateInit2(&strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15 + 16, 8, Z_DEFAULT_STRATEGY);
-  if (zret != Z_OK)
+  // Set up a gzip stream (the "+ 16" in 15 + 16), instead of a zlib stream (default)
+  ret = deflateInit2(&strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 15 + 16, 8, Z_DEFAULT_STRATEGY);
+  if (ret != Z_OK)
     {
-      DPRINTF(E_DBG, L_HTTPD, "zlib setup failed: %s\n", zError(zret));
-
-      goto out_fail_init;
+      DPRINTF(E_LOG, L_HTTPD, "zlib setup failed: %s\n", zError(ret));
+      return NULL;
     }
 
-  strm.next_in = EVBUFFER_DATA(evbuf);
-  strm.avail_in = EVBUFFER_LENGTH(evbuf);
+  strm.next_in = evbuffer_pullup(in, -1);
+  strm.avail_in = evbuffer_get_length(in);
 
-  flush = Z_NO_FLUSH;
-
-  /* 2 iterations: Z_NO_FLUSH until input is consumed, then Z_FINISH */
-  for (;;)
+  out = evbuffer_new();
+  if (!out)
     {
-      do
-	{
-	  strm.next_out = outbuf;
-	  strm.avail_out = sizeof(outbuf);
-
-	  zret = deflate(&strm, flush);
-	  if (zret == Z_STREAM_ERROR)
-	    {
-	      DPRINTF(E_LOG, L_HTTPD, "Could not deflate data: %s\n", strm.msg);
-
-	      goto out_fail_gz;
-	    }
-
-	  ret = evbuffer_add(gzbuf, outbuf, sizeof(outbuf) - strm.avail_out);
-	  if (ret < 0)
-	    {
-	      DPRINTF(E_LOG, L_HTTPD, "Out of memory adding gzipped data to evbuffer\n");
-
-	      goto out_fail_gz;
-	    }
-	}
-      while (strm.avail_out == 0);
-
-      if (flush == Z_FINISH)
-	break;
-
-      flush = Z_FINISH;
+      DPRINTF(E_LOG, L_HTTPD, "Could not allocate evbuffer for gzipped reply\n");
+      goto out_deflate_end;
     }
 
-  if (zret != Z_STREAM_END)
+  // We use this to avoid a memcpy. The 512 is an arbitrary padding to make sure
+  // there is enough space, even if the compressed output should be slightly
+  // larger than input (could happen with small inputs).
+  ret = evbuffer_reserve_space(out, strm.avail_in + 512, iovec, 1);
+  if (ret < 0)
     {
-      DPRINTF(E_LOG, L_HTTPD, "Compressed data not finalized!\n");
-
-      goto out_fail_gz;
+      DPRINTF(E_LOG, L_HTTPD, "Could not reserve memory for gzipped reply\n");
+      goto out_evbuf_free;
     }
 
+  strm.next_out = iovec[0].iov_base;
+  strm.avail_out = iovec[0].iov_len;
+
+  ret = deflate(&strm, Z_FINISH);
+  if (ret != Z_STREAM_END)
+    goto out_evbuf_free;
+
+  iovec[0].iov_len -= strm.avail_out;
+
+  evbuffer_commit_space(out, iovec, 1);
   deflateEnd(&strm);
 
-  headers = evhttp_request_get_output_headers(req);
+  return out;
 
-  evhttp_add_header(headers, "Content-Encoding", "gzip");
-  evhttp_send_reply(req, code, reason, gzbuf);
+ out_evbuf_free:
+  evbuffer_free(out);
 
-  evbuffer_free(gzbuf);
-
-  /* Drain original buffer, as would be after evhttp_send_reply() */
-  evbuffer_drain(evbuf, EVBUFFER_LENGTH(evbuf));
-
-  return;
-
- out_fail_gz:
+ out_deflate_end:
   deflateEnd(&strm);
- out_fail_init:
-  evbuffer_free(gzbuf);
- no_gzip:
-  evhttp_send_reply(req, code, reason, evbuf);
+
+  return NULL;
+}
+
+void
+httpd_send_reply(struct evhttp_request *req, int code, const char *reason, struct evbuffer *evbuf, enum httpd_send_flags flags)
+{
+  struct evbuffer *gzbuf;
+  struct evkeyvalq *input_headers;
+  struct evkeyvalq *output_headers;
+  const char *param;
+  int do_gzip;
+
+  if (!req)
+    return;
+
+  input_headers = evhttp_request_get_input_headers(req);
+  output_headers = evhttp_request_get_output_headers(req);
+
+  do_gzip = ( (!(flags & HTTPD_SEND_NO_GZIP)) &&
+              evbuf && (evbuffer_get_length(evbuf) > 512) &&
+              (param = evhttp_find_header(input_headers, "Accept-Encoding")) &&
+              (strstr(param, "gzip") || strstr(param, "*"))
+            );
+
+  if (allow_origin)
+    evhttp_add_header(output_headers, "Access-Control-Allow-Origin", allow_origin);
+
+  if (do_gzip && (gzbuf = httpd_gzip_deflate(evbuf)))
+    {
+      DPRINTF(E_DBG, L_HTTPD, "Gzipping response\n");
+
+      evhttp_add_header(output_headers, "Content-Encoding", "gzip");
+      evhttp_send_reply(req, code, reason, gzbuf);
+      evbuffer_free(gzbuf);
+
+      // Drain original buffer, as would be after evhttp_send_reply()
+      evbuffer_drain(evbuf, evbuffer_get_length(evbuf));
+    }
+  else
+    {
+      evhttp_send_reply(req, code, reason, evbuf);
+    }
+}
+
+// This is a modified version of evhttp_send_error (credit libevent)
+void
+httpd_send_error(struct evhttp_request* req, int error, const char* reason)
+{
+  struct evkeyvalq *output_headers;
+  struct evbuffer *evbuf;
+
+  if (!allow_origin)
+    {
+      evhttp_send_error(req, error, reason);
+      return;
+    }
+
+  output_headers = evhttp_request_get_output_headers(req);
+
+  evhttp_clear_headers(output_headers);
+
+  evhttp_add_header(output_headers, "Access-Control-Allow-Origin", allow_origin);
+  evhttp_add_header(output_headers, "Content-Type", "text/html");
+  evhttp_add_header(output_headers, "Connection", "close");
+
+  evbuf = evbuffer_new();
+  if (!evbuf)
+    DPRINTF(E_LOG, L_HTTPD, "Could not allocate evbuffer for error page\n");
+  else
+    evbuffer_add_printf(evbuf, ERR_PAGE, error, reason, reason);
+
+  evhttp_send_reply(req, error, reason, evbuf);
+
+  if (evbuf)
+    evbuffer_free(evbuf);
 }
 
 /* Thread: httpd */
@@ -807,14 +860,14 @@ redirect_to_index(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Redirection URL exceeds buffer length\n");
 
-      evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+      httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
       return;
     }
 
   headers = evhttp_request_get_output_headers(req);
-
   evhttp_add_header(headers, "Location", buf);
-  evhttp_send_reply(req, HTTP_MOVETEMP, "Moved", NULL);
+
+  httpd_send_reply(req, HTTP_MOVETEMP, "Moved", NULL, HTTPD_SEND_NO_GZIP);
 }
 
 /* Thread: httpd */
@@ -854,7 +907,7 @@ serve_file(struct evhttp_request *req, char *uri)
 	{
 	  DPRINTF(E_LOG, L_HTTPD, "Remote web interface request denied; no password set\n");
 
-	  evhttp_send_error(req, 403, "Forbidden");
+	  httpd_send_error(req, 403, "Forbidden");
 	  return;
 	}
     }
@@ -864,7 +917,7 @@ serve_file(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Request exceeds PATH_MAX: %s\n", uri);
 
-      evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+      httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
 
       return;
     }
@@ -874,7 +927,7 @@ serve_file(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not lstat() %s: %s\n", path, strerror(errno));
 
-      evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+      httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
 
       return;
     }
@@ -892,7 +945,7 @@ serve_file(struct evhttp_request *req, char *uri)
 	{
 	  DPRINTF(E_LOG, L_HTTPD, "Could not dereference %s: %s\n", path, strerror(errno));
 
-	  evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+	  httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
 
 	  return;
 	}
@@ -901,7 +954,7 @@ serve_file(struct evhttp_request *req, char *uri)
 	{
 	  DPRINTF(E_LOG, L_HTTPD, "Dereferenced path exceeds PATH_MAX: %s\n", path);
 
-	  evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+	  httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
 
 	  free(deref);
 	  return;
@@ -915,7 +968,7 @@ serve_file(struct evhttp_request *req, char *uri)
 	{
 	  DPRINTF(E_LOG, L_HTTPD, "Could not stat() %s: %s\n", path, strerror(errno));
 
-	  evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+	  httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
 
 	  return;
 	}
@@ -930,7 +983,7 @@ serve_file(struct evhttp_request *req, char *uri)
 
   if (path_is_legal(path) != 0)
     {
-      evhttp_send_error(req, 403, "Forbidden");
+      httpd_send_error(req, 403, "Forbidden");
 
       return;
     }
@@ -940,7 +993,7 @@ serve_file(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not create evbuffer\n");
 
-      evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal error");
+      httpd_send_error(req, HTTP_SERVUNAVAIL, "Internal error");
       return;
     }
 
@@ -949,7 +1002,7 @@ serve_file(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not open %s: %s\n", path, strerror(errno));
 
-      evhttp_send_error(req, HTTP_NOTFOUND, "Not Found");
+      httpd_send_error(req, HTTP_NOTFOUND, "Not Found");
       return;
     }
 
@@ -962,7 +1015,7 @@ serve_file(struct evhttp_request *req, char *uri)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not read file into evbuffer\n");
 
-      evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal error");
+      httpd_send_error(req, HTTP_SERVUNAVAIL, "Internal error");
       return;
     }
 
@@ -981,9 +1034,9 @@ serve_file(struct evhttp_request *req, char *uri)
     }
 
   headers = evhttp_request_get_output_headers(req);
-
   evhttp_add_header(headers, "Content-Type", ctype);
-  evhttp_send_reply(req, HTTP_OK, "OK", evbuf);
+
+  httpd_send_reply(req, HTTP_OK, "OK", evbuf, HTTPD_SEND_NO_GZIP);
 
   evbuffer_free(evbuf);
 }
@@ -992,10 +1045,32 @@ serve_file(struct evhttp_request *req, char *uri)
 static void
 httpd_gen_cb(struct evhttp_request *req, void *arg)
 {
+  struct evkeyvalq *input_headers;
+  struct evkeyvalq *output_headers;
   const char *req_uri;
   char *uri;
   char *ptr;
 
+  // Did we get a CORS preflight request?
+  input_headers = evhttp_request_get_input_headers(req);
+  if ( input_headers && allow_origin &&
+       (evhttp_request_get_command(req) == EVHTTP_REQ_OPTIONS) &&
+       evhttp_find_header(input_headers, "Origin") &&
+       evhttp_find_header(input_headers, "Access-Control-Request-Method") )
+    {
+      output_headers = evhttp_request_get_output_headers(req);
+
+      evhttp_add_header(output_headers, "Access-Control-Allow-Origin", allow_origin);
+
+      // Allow only GET method and authorization header in cross origin requests
+      evhttp_add_header(output_headers, "Access-Control-Allow-Method", "GET");
+      evhttp_add_header(output_headers, "Access-Control-Allow-Headers", "authorization");
+
+      // In this case there is no reason to go through httpd_send_reply
+      evhttp_send_reply(req, HTTP_OK, "OK", NULL);
+      return;
+    }
+
   req_uri = evhttp_request_get_uri(req);
   if (!req_uri)
     {
@@ -1034,6 +1109,12 @@ httpd_gen_cb(struct evhttp_request *req, void *arg)
     {
       dacp_request(req);
 
+      goto out;
+    }
+  else if (streaming_is_request(req, uri))
+    {
+      streaming_request(req);
+
       goto out;
     }
 
@@ -1240,28 +1321,30 @@ httpd_basic_auth(struct evhttp_request *req, char *user, char *passwd, char *rea
   header = (char *)malloc(len);
   if (!header)
     {
-      evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
+      httpd_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
       return -1;
     }
 
   ret = snprintf(header, len, "Basic realm=\"%s\"", realm);
   if ((ret < 0) || (ret >= len))
     {
-      evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
+      httpd_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
       return -1;
     }
 
   evbuf = evbuffer_new();
   if (!evbuf)
     {
-      evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
+      httpd_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
       return -1;
     }
 
   headers = evhttp_request_get_output_headers(req);
   evhttp_add_header(headers, "WWW-Authenticate", header);
+
   evbuffer_add(evbuf, http_reply_401, strlen(http_reply_401));
-  evhttp_send_reply(req, 401, "Unauthorized", evbuf);
+
+  httpd_send_reply(req, 401, "Unauthorized", evbuf, HTTPD_SEND_NO_GZIP);
 
   free(header);
   evbuffer_free(evbuf);
@@ -1273,7 +1356,7 @@ httpd_basic_auth(struct evhttp_request *req, char *user, char *passwd, char *rea
 int
 httpd_init(void)
 {
-  const char *addr;
+  int v6enabled;
   unsigned short port;
   int ret;
 
@@ -1311,6 +1394,8 @@ httpd_init(void)
       goto dacp_fail;
     }
 
+  streaming_init();
+
 #ifdef USE_EVENTFD
   exit_efd = eventfd(0, EFD_CLOEXEC);
   if (exit_efd < 0)
@@ -1319,8 +1404,10 @@ httpd_init(void)
 
       goto pipe_fail;
     }
+
+  exitev = event_new(evbase_httpd, exit_efd, EV_READ, exit_cb, NULL);
 #else
-# if defined(__linux__)
+# ifdef HAVE_PIPE2
   ret = pipe2(exit_pipe, O_CLOEXEC);
 # else
   ret = pipe(exit_pipe);
@@ -1331,37 +1418,56 @@ httpd_init(void)
 
       goto pipe_fail;
     }
-#endif /* USE_EVENTFD */
 
-#ifdef USE_EVENTFD
-  event_set(&exitev, exit_efd, EV_READ, exit_cb, NULL);
-#else
-  event_set(&exitev, exit_pipe[0], EV_READ, exit_cb, NULL);
-#endif
-  event_base_set(evbase_httpd, &exitev);
-  event_add(&exitev, NULL);
+  exitev = event_new(evbase_httpd, exit_pipe[0], EV_READ, exit_cb, NULL);
+#endif /* USE_EVENTFD */
+  if (!exitev)
+    {
+      DPRINTF(E_FATAL, L_HTTPD, "Could not create exit event\n");
+
+      goto event_fail;
+    }
+  event_add(exitev, NULL);
 
   evhttpd = evhttp_new(evbase_httpd);
   if (!evhttpd)
     {
       DPRINTF(E_FATAL, L_HTTPD, "Could not create HTTP server\n");
 
-      goto evhttp_fail;
+      goto event_fail;
     }
 
-  if (cfg_getbool(cfg_getsec(cfg, "general"), "ipv6"))
-    addr = "::";
-  else
-    addr = "0.0.0.0";
-
+  v6enabled = cfg_getbool(cfg_getsec(cfg, "general"), "ipv6");
   port = cfg_getint(cfg_getsec(cfg, "library"), "port");
 
-  ret = evhttp_bind_socket(evhttpd, addr, port);
-  if (ret < 0)
+  // For CORS headers
+  allow_origin = cfg_getstr(cfg_getsec(cfg, "general"), "allow_origin");
+  if (allow_origin)
     {
-      DPRINTF(E_FATAL, L_HTTPD, "Could not bind %s:%d\n", addr, port);
+      if (strlen(allow_origin) != 0)
+	evhttp_set_allowed_methods(evhttpd, EVHTTP_REQ_GET | EVHTTP_REQ_POST | EVHTTP_REQ_HEAD | EVHTTP_REQ_OPTIONS);
+      else
+	allow_origin = NULL;
+    }
 
-      goto bind_fail;
+  if (v6enabled)
+    {
+      ret = evhttp_bind_socket(evhttpd, "::", port);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_HTTPD, "Could not bind to port %d with IPv6, falling back to IPv4\n", port);
+	  v6enabled = 0;
+	}
+    }
+
+  if (!v6enabled)
+    {
+      ret = evhttp_bind_socket(evhttpd, "0.0.0.0", port);
+      if (ret < 0)
+	{
+	  DPRINTF(E_FATAL, L_HTTPD, "Could not bind to port %d (forked-daapd already running?)\n", port);
+	  goto bind_fail;
+	}
     }
 
   evhttp_set_gencb(evhttpd, httpd_gen_cb, NULL);
@@ -1374,12 +1480,18 @@ httpd_init(void)
       goto thread_fail;
     }
 
+#if defined(HAVE_PTHREAD_SETNAME_NP)
+  pthread_setname_np(tid_httpd, "httpd");
+#elif defined(HAVE_PTHREAD_SET_NAME_NP)
+  pthread_set_name_np(tid_httpd, "httpd");
+#endif
+
   return 0;
 
  thread_fail:
  bind_fail:
   evhttp_free(evhttpd);
- evhttp_fail:
+ event_fail:
 #ifdef USE_EVENTFD
   close(exit_efd);
 #else
@@ -1387,6 +1499,7 @@ httpd_init(void)
   close(exit_pipe[1]);
 #endif
  pipe_fail:
+  streaming_deinit();
   dacp_deinit();
  dacp_fail:
   daap_deinit();
@@ -1432,6 +1545,7 @@ httpd_deinit(void)
       return;
     }
 
+  streaming_deinit();
   rsp_deinit();
   dacp_deinit();
   daap_deinit();
