@@ -47,9 +47,10 @@
 #include <pthread.h>
 
 #include <getopt.h>
-#include <event.h>
+#include <event2/event.h>
 #include <libavutil/log.h>
 #include <libavformat/avformat.h>
+#include <libavfilter/avfilter.h>
 
 #include <gcrypt.h>
 GCRY_THREAD_OPTION_PTHREAD_IMPL;
@@ -78,14 +79,14 @@ GCRY_THREAD_OPTION_PTHREAD_IMPL;
 
 struct event_base *evbase_main;
 
-static struct event sig_event;
+static struct event *sig_event;
 static int main_exit;
 
 static void
 version(void)
 {
   fprintf(stdout, "Forked Media Server: Version %s\n", VERSION);
-  fprintf(stdout, "Copyright (C) 2009-2011 Julien BLACHE <jb@jblache.org>\n");
+  fprintf(stdout, "Copyright (C) 2009-2015 Julien BLACHE <jb@jblache.org>\n");
   fprintf(stdout, "Based on mt-daapd, Copyright (C) 2003-2007 Ron Pedde <ron@pedde.com>\n");
   fprintf(stdout, "Released under the GNU General Public License version 2 or later\n");
 }
@@ -220,14 +221,16 @@ daemonize(int background, char *pidfile)
 }
 
 static int
-register_services(char *ffid, int no_rsp, int no_daap)
+register_services(char *ffid, int no_rsp, int no_daap, int mdns_no_mpd)
 {
   cfg_t *lib;
+  cfg_t *mpd;
   char *libname;
   char *password;
   char *txtrecord[10];
   char records[9][128];
   int port;
+  int mpd_port;
   uint32_t hash;
   int i;
   int ret;
@@ -317,6 +320,16 @@ register_services(char *ffid, int no_rsp, int no_daap)
   if (ret < 0)
     return ret;
 
+  /* Register MPD serivce */
+  mpd = cfg_getsec(cfg, "mpd");
+  mpd_port = cfg_getint(mpd, "port");
+  if (!mdns_no_mpd && mpd_port > 0)
+    {
+      ret = mdns_register(libname, "_mpd._tcp", mpd_port, NULL);
+            if (ret < 0)
+      	return ret;
+    }
+
   return 0;
 }
 
@@ -333,9 +346,9 @@ signal_signalfd_cb(int fd, short event, void *arg)
       switch (info.ssi_signo)
 	{
 	  case SIGCHLD:
-	    DPRINTF(E_LOG, L_MAIN, "Got SIGCHLD, reaping children\n");
+	    DPRINTF(E_LOG, L_MAIN, "Got SIGCHLD\n");
 
-	    while (wait3(&status, WNOHANG, NULL) > 0)
+	    while (waitpid(-1, &status, WNOHANG) > 0)
 	      /* Nothing. */ ;
 	    break;
 
@@ -358,7 +371,7 @@ signal_signalfd_cb(int fd, short event, void *arg)
   if (main_exit)
     event_base_loopbreak(evbase_main);
   else
-    event_add(&sig_event, NULL);
+    event_add(sig_event, NULL);
 }
 
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
@@ -378,9 +391,9 @@ signal_kqueue_cb(int fd, short event, void *arg)
       switch (ke.ident)
 	{
 	  case SIGCHLD:
-	    DPRINTF(E_LOG, L_MAIN, "Got SIGCHLD, reaping children\n");
+	    DPRINTF(E_LOG, L_MAIN, "Got SIGCHLD\n");
 
-	    while (wait3(&status, WNOHANG, NULL) > 0)
+	    while (waitpid(-1, &status, WNOHANG) > 0)
 	      /* Nothing. */ ;
 	    break;
 
@@ -403,7 +416,7 @@ signal_kqueue_cb(int fd, short event, void *arg)
   if (main_exit)
     event_base_loopbreak(evbase_main);
   else
-    event_add(&sig_event, NULL);
+    event_add(sig_event, NULL);
 }
 #endif
 
@@ -445,11 +458,13 @@ main(int argc, char **argv)
   int background;
   int mdns_no_rsp;
   int mdns_no_daap;
+  int mdns_no_mpd;
   int loglevel;
   char *logdomains;
   char *logfile;
   char *ffid;
   char *pidfile;
+  char buildopts[256];
   const char *gcry_version;
   sigset_t sigs;
   int sigfd;
@@ -483,6 +498,7 @@ main(int argc, char **argv)
   ffid = NULL;
   mdns_no_rsp = 0;
   mdns_no_daap = 0;
+  mdns_no_mpd = 1; // only announce if mpd protocol support is activated
 
   while ((option = getopt_long(argc, argv, "D:d:c:P:fb:v", option_map, NULL)) != -1)
     {
@@ -575,6 +591,32 @@ main(int argc, char **argv)
 
   DPRINTF(E_LOG, L_MAIN, "Forked Media Server Version %s taking off\n", VERSION);
 
+  /* Remember to check the size of buildopts when adding new opts */
+  strcpy(buildopts, "");
+#ifdef ITUNES
+  strcat(buildopts, " --enable-itunes");
+#endif
+#ifdef SPOTIFY
+  strcat(buildopts, " --enable-spotify");
+#endif
+#ifdef LASTFM
+  strcat(buildopts, " --enable-lastfm");
+#endif
+#ifdef CHROMECAST
+  strcat(buildopts, " --enable-chromecast");
+#endif
+#ifdef MPD
+  strcat(buildopts, " --enable-mpd");
+#endif
+#ifdef ALSA
+  strcat(buildopts, " --with-alsa");
+#endif
+#ifdef PULSEAUDIO
+  strcat(buildopts, " --with-pulseaudio");
+#endif
+
+  DPRINTF(E_LOG, L_MAIN, "Built %s with:%s\n", BUILDDATE, buildopts);
+
   ret = av_lockmgr_register(ffmpeg_lockmgr);
   if (ret < 0)
     {
@@ -585,6 +627,7 @@ main(int argc, char **argv)
     }
 
   av_register_all();
+  avfilter_register_all();
 #if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 13)
   avformat_network_init();
 #endif
@@ -642,8 +685,8 @@ main(int argc, char **argv)
       goto daemon_fail;
     }
 
-  /* Initialize libevent (after forking) */
-  evbase_main = event_init();
+  /* Initialize event base (after forking) */
+  evbase_main = event_base_new();
 
   DPRINTF(E_LOG, L_MAIN, "mDNS init\n");
   ret = mdns_init();
@@ -745,6 +788,7 @@ main(int argc, char **argv)
       ret = EXIT_FAILURE;
       goto mpd_fail;
     }
+  mdns_no_mpd = 0;
 #endif
 
   /* Start Remote pairing service */
@@ -758,7 +802,7 @@ main(int argc, char **argv)
     }
 
   /* Register mDNS services */
-  ret = register_services(ffid, mdns_no_rsp, mdns_no_daap);
+  ret = register_services(ffid, mdns_no_rsp, mdns_no_daap, mdns_no_mpd);
   if (ret < 0)
     {
       ret = EXIT_FAILURE;
@@ -776,8 +820,7 @@ main(int argc, char **argv)
       goto signalfd_fail;
     }
 
-  event_set(&sig_event, sigfd, EV_READ, signal_signalfd_cb, NULL);
-
+  sig_event = event_new(evbase_main, sigfd, EV_READ, signal_signalfd_cb, NULL);
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
   sigfd = kqueue();
   if (sigfd < 0)
@@ -802,11 +845,17 @@ main(int argc, char **argv)
       goto signalfd_fail;
     }
 
-  event_set(&sig_event, sigfd, EV_READ, signal_kqueue_cb, NULL);
+  sig_event = event_new(evbase_main, sigfd, EV_READ, signal_kqueue_cb, NULL);
 #endif
+  if (!sig_event)
+    {
+      DPRINTF(E_FATAL, L_MAIN, "Could not create signal event\n");
 
-  event_base_set(evbase_main, &sig_event);
-  event_add(&sig_event, NULL);
+      ret = EXIT_FAILURE;
+      goto sig_event_fail;
+    }
+
+  event_add(sig_event, NULL);
 
   /* Run the loop */
   event_base_dispatch(evbase_main);
@@ -821,6 +870,7 @@ main(int argc, char **argv)
   DPRINTF(E_LOG, L_MAIN, "mDNS deinit\n");
   mdns_deinit();
 
+ sig_event_fail:
  signalfd_fail:
  mdns_reg_fail:
   DPRINTF(E_LOG, L_MAIN, "Remote pairing deinit\n");
