@@ -110,9 +110,9 @@ typedef void (*cast_reply_cb)(struct cast_session *cs, struct cast_msg_payload *
 enum cast_state
 {
   // Something bad happened during a session
-  CAST_STATE_FAILED          = -1,
+  CAST_STATE_FAILED          = 0,
   // No session allocated
-  CAST_STATE_NULL            = 0,
+  CAST_STATE_NONE            = 1,
   // Session allocated, but no connection
   CAST_STATE_DISCONNECTED    = CAST_STATE_F_STARTUP | 0x01,
   // TCP connect, TLS handshake, CONNECT and GET_STATUS request
@@ -162,6 +162,11 @@ struct cast_session
   cast_reply_cb callback_register[CALLBACK_REGISTER_SIZE];
   struct event *reply_timeout;
 
+  // This is used to work around a bug where no response is given by the device.
+  // For certain requests, we will then retry, e.g. by checking status. We
+  // register our retry so that we on only retry once.
+  int retry;
+
   // Session info from the ChromeCast
   char *transport_id;
   char *session_id;
@@ -604,7 +609,7 @@ cast_msg_send(struct cast_session *cs, enum cast_msg_types type, cast_reply_cb r
   len = extensions__core_api__cast_channel__cast_message__get_packed_size(&msg);
   if (len <= 0)
     {
-      DPRINTF(E_LOG, L_CAST, "Could not send message (type %d), invalid length: %d\n", type, len);
+      DPRINTF(E_LOG, L_CAST, "Could not send message (type %d), invalid length: %zu\n", type, len);
       return -1;
     }
 
@@ -628,7 +633,7 @@ cast_msg_send(struct cast_session *cs, enum cast_msg_types type, cast_reply_cb r
     }
 
   if (type != PONG)
-    DPRINTF(E_DBG, L_CAST, "TX %d %s %s %s %s\n", len, msg.source_id, msg.destination_id, msg.namespace_, msg.payload_utf8);
+    DPRINTF(E_DBG, L_CAST, "TX %zu %s %s %s %s\n", len, msg.source_id, msg.destination_id, msg.namespace_, msg.payload_utf8);
 
   return 0;
 }
@@ -728,6 +733,15 @@ cast_msg_process(struct cast_session *cs, const uint8_t *data, size_t len)
   int unknown_session_id;
   int i;
 
+#ifdef DEBUG_CONNECTION
+  char *b64 = b64_encode(data, len);
+  if (b64)
+    {
+      DPRINTF(E_DBG, L_CAST, "Reply dump (len %zu): %s\n", len, b64);
+      free(b64);
+    }
+#endif
+
   reply = extensions__core_api__cast_channel__cast_message__unpack(NULL, len, data);
   if (!reply)
     {
@@ -748,7 +762,7 @@ cast_msg_process(struct cast_session *cs, const uint8_t *data, size_t len)
       goto out_free_parsed;
     }
 
-  DPRINTF(E_DBG, L_CAST, "RX %d %s %s %s %s\n", len, reply->source_id, reply->destination_id, reply->namespace_, reply->payload_utf8);
+  DPRINTF(E_DBG, L_CAST, "RX %zu %s %s %s %s\n", len, reply->source_id, reply->destination_id, reply->namespace_, reply->payload_utf8);
 
   if (payload.type == UNKNOWN)
     goto out_free_parsed;
@@ -795,7 +809,7 @@ cast_msg_process(struct cast_session *cs, const uint8_t *data, size_t len)
 
 /*	  cs->state = CAST_STATE_MEDIA_CONNECTED;
 	  // Kill the session, the player will need to restart it
-	  cast_session_shutdown(cs, CAST_STATE_NULL);
+	  cast_session_shutdown(cs, CAST_STATE_NONE);
 	  goto out_free_parsed;
 */	}
     }
@@ -823,7 +837,7 @@ cast_status(struct cast_session *cs)
       case CAST_STATE_FAILED:
 	state = OUTPUT_STATE_FAILED;
 	break;
-      case CAST_STATE_NULL:
+      case CAST_STATE_NONE:
 	state = OUTPUT_STATE_STOPPED;
 	break;
       case CAST_STATE_DISCONNECTED ... CAST_STATE_MEDIA_LAUNCHED:
@@ -925,17 +939,34 @@ cast_cb_startup_launch(struct cast_session *cs, struct cast_msg_payload *payload
 {
   int ret;
 
+  // Sometimes the response to a LAUNCH is just a broadcast RECEIVER_STATUS
+  // without our requestId. That won't be registered by our response handler,
+  // and we get an empty callback due to timeout. In this case we send a
+  // GET_STATUS to see if we are good to go anyway.
+  if (!payload && !cs->retry)
+    {
+      DPRINTF(E_LOG, L_CAST, "No RECEIVER_STATUS reply to our LAUNCH - trying GET_STATUS instead\n");
+      cs->retry++;
+      ret = cast_msg_send(cs, GET_STATUS, cast_cb_startup_launch);
+      if (ret != 0)
+	goto error;
+
+      return;
+    }
+
   if (!payload)
     {
       DPRINTF(E_LOG, L_CAST, "No RECEIVER_STATUS reply to our LAUNCH - aborting\n");
       goto error;
     }
-  else if (payload->type != RECEIVER_STATUS)
+
+  if (payload->type != RECEIVER_STATUS)
     {
       DPRINTF(E_LOG, L_CAST, "No RECEIVER_STATUS reply to our LAUNCH (got type: %d) - aborting\n", payload->type);
       goto error;
     }
-  else if (!payload->transport_id || !payload->session_id)
+
+  if (!payload->transport_id || !payload->session_id)
     {
       DPRINTF(E_LOG, L_CAST, "Missing session id or transport id in RECEIVER_STATUS - aborting\n");
       goto error;
@@ -947,6 +978,8 @@ cast_cb_startup_launch(struct cast_session *cs, struct cast_msg_payload *payload
   cs->session_id = strdup(payload->session_id);
   cs->transport_id = strdup(payload->transport_id);
 
+  cs->retry = 0;
+
   ret = cast_msg_send(cs, MEDIA_CONNECT, NULL);
   if (ret == 0)
     ret = cast_msg_send(cs, MEDIA_GET_STATUS, cast_cb_startup_media);
@@ -1009,7 +1042,7 @@ cast_cb_probe(struct cast_session *cs, struct cast_msg_payload *payload)
 
   cast_status(cs);
 
-  cast_session_shutdown(cs, CAST_STATE_NULL);
+  cast_session_shutdown(cs, CAST_STATE_NONE);
 
   return;
 
@@ -1078,66 +1111,91 @@ cast_listen_cb(int fd, short what, void *arg)
 {
   struct cast_session *cs;
   uint8_t buffer[MAX_BUF + 1]; // Not sure about the +1, but is copied from gnutls examples
+  uint32_t be;
+  size_t len;
   int received;
   int ret;
 
-  cs = (struct cast_session *)arg;
+  for (cs = sessions; cs; cs = cs->next)
+    {
+      if (cs == (struct cast_session *)arg)
+	break;
+    }
+
+  if (!cs)
+    {
+      DPRINTF(E_INFO, L_CAST, "Callback on dead session, ignoring\n");
+      return;
+    }
 
   if (what == EV_TIMEOUT)
     {
       DPRINTF(E_LOG, L_CAST, "No heartbeat from '%s', shutting down\n", cs->devname);
-      cs->state = CAST_STATE_CONNECTED;
-      cast_session_shutdown(cs, CAST_STATE_FAILED);
-      return;
+      goto fail;
     }
 
 #ifdef DEBUG_CONNECTION
   DPRINTF(E_DBG, L_CAST, "New data from '%s'\n", cs->devname);
 #endif
 
-  received = 0;
-  while ((ret = gnutls_record_recv(cs->tls_session, buffer + received, MAX_BUF - received)) > 0)
+  // We first read the 4 byte header and then the actual message. The header
+  // will be the length of the message.
+  ret = gnutls_record_recv(cs->tls_session, buffer, 4);
+  if (ret != 4)
+    goto no_read;
+
+  memcpy(&be, buffer, 4);
+  len = be32toh(be);
+  if ((len == 0) || (len > MAX_BUF))
     {
-#ifdef DEBUG_CONNECTION
-      DPRINTF(E_DBG, L_CAST, "Received %d bytes\n", ret);
-#endif
-
-      if (ret == 4)
-	{
-#ifdef DEBUG_CONNECTION
-	  uint32_t be;
-	  size_t len;
-	  memcpy(&be, buffer, 4);
-	  len = be32toh(be);
-	  DPRINTF(E_DBG, L_CAST, "Incoming %d bytes\n", len);
-#endif
-	}
-      else
-	{
-	  received += ret;
-	}
-
-      if (received >= MAX_BUF)
-	{
-	  DPRINTF(E_LOG, L_CAST, "Receive buffer exhausted!\n");
-	  cast_session_shutdown(cs, CAST_STATE_FAILED);
-	  return;
-	}
+      DPRINTF(E_LOG, L_CAST, "Bad length of incoming message, aborting (len=%zu, size=%d)\n", len, MAX_BUF);
+      goto fail;
     }
 
+  received = 0;
+  while (received < len)
+    {
+      ret = gnutls_record_recv(cs->tls_session, buffer + received, len - received);
+      if (ret <= 0)
+	goto no_read;
+
+      received += ret;
+
+#ifdef DEBUG_CONNECTION
+      DPRINTF(E_DBG, L_CAST, "Received %d bytes out of expected %zu bytes\n", received, len);
+#endif
+    }
+
+  ret = gnutls_record_check_pending(cs->tls_session);
+
+  // Process the message - note that this may result in cs being invalidated
+  cast_msg_process(cs, buffer, len);
+
+  // In the event there was more data waiting for us we go again
+  if (ret > 0)
+    {
+      DPRINTF(E_INFO, L_CAST, "More data pending from device (%d bytes)\n", ret);
+      cast_listen_cb(fd, what, arg);
+    }
+
+  return;
+
+ no_read:
   if ((ret != GNUTLS_E_INTERRUPTED) && (ret != GNUTLS_E_AGAIN))
     {
       DPRINTF(E_LOG, L_CAST, "Session error: %s\n", gnutls_strerror(ret));
-
-      // Downgrade state to make cast_session_shutdown perform an exit which is
-      // quick and won't require a reponse from the device
-      cs->state = CAST_STATE_CONNECTED;
-      cast_session_shutdown(cs, CAST_STATE_FAILED);
-      return;
+      goto fail;
     }
 
-  if (received)
-    cast_msg_process(cs, buffer, received);
+  DPRINTF(E_DBG, L_CAST, "Return value from tls is %d (GNUTLS_E_AGAIN is %d)\n", ret, GNUTLS_E_AGAIN);
+
+  return;
+
+ fail:
+  // Downgrade state to make cast_session_shutdown perform an exit which is
+  // quick and won't require a reponse from the device
+  cs->state = CAST_STATE_CONNECTED;
+  cast_session_shutdown(cs, CAST_STATE_FAILED);
 }
 
 static void
@@ -1147,15 +1205,15 @@ cast_reply_timeout_cb(int fd, short what, void *arg)
   int i;
 
   cs = (struct cast_session *)arg;
+  i = cs->request_id % CALLBACK_REGISTER_SIZE;
 
-  DPRINTF(E_WARN, L_CAST, "Request timeout, will run empty callbacks\n");
+  DPRINTF(E_LOG, L_CAST, "Request %d timed out, will run empty callback\n", i);
 
-  for (i = 0; i < CALLBACK_REGISTER_SIZE; i++)
-    if (cs->callback_register[i])
-      {
-	cs->callback_register[i](cs, NULL);
-	cs->callback_register[i] = NULL;
-      }
+  if (cs->callback_register[i])
+    {
+      cs->callback_register[i](cs, NULL);
+      cs->callback_register[i] = NULL;
+    }
 }
 
 static void
@@ -1358,7 +1416,7 @@ cast_session_make(struct output_device *device, int family, output_status_cb cb)
 }
 
 // Attempts to "nicely" bring down a session to wanted_state, and then issues
-// the callback. If wanted_state is CAST_STATE_NULL/FAILED then the session is purged.
+// the callback. If wanted_state is CAST_STATE_NONE/FAILED then the session is purged.
 static void
 cast_session_shutdown(struct cast_session *cs, enum cast_state wanted_state)
 {
@@ -1372,7 +1430,7 @@ cast_session_shutdown(struct cast_session *cs, enum cast_state wanted_state)
     }
   else if (cs->state < wanted_state)
     {
-      DPRINTF(E_LOG, L_CAST, "Bug! Shutdown request got wanted_state that is higher than current state\n");
+      DPRINTF(E_LOG, L_CAST, "Bug! Shutdown request got wanted_state (%d) that is higher than current state (%d)\n", wanted_state, cs->state);
       return;
     }
 
@@ -1429,7 +1487,7 @@ cast_session_shutdown(struct cast_session *cs, enum cast_state wanted_state)
     return;
 
   // Asked to destroy the session
-  if (wanted_state == CAST_STATE_NULL || wanted_state == CAST_STATE_FAILED)
+  if (wanted_state == CAST_STATE_NONE || wanted_state == CAST_STATE_FAILED)
     {
       cs->state = wanted_state;
       cast_status(cs);
@@ -1488,7 +1546,7 @@ cast_device_stop(struct output_session *session)
 {
   struct cast_session *cs = session->session;
 
-  cast_session_shutdown(cs, CAST_STATE_NULL);
+  cast_session_shutdown(cs, CAST_STATE_NONE);
 }
 
 static int
@@ -1580,9 +1638,14 @@ static void
 cast_playback_stop(void)
 {
   struct cast_session *cs;
+  struct cast_session *next;
 
-  for (cs = sessions; cs; cs = cs->next)
-    cast_session_shutdown(cs, CAST_STATE_NULL);
+  for (cs = sessions; cs; cs = next)
+    {
+      next = cs->next;
+      if (cs->state & CAST_STATE_F_MEDIA_CONNECTED)
+	cast_session_shutdown(cs, CAST_STATE_NONE);
+    }
 }
 
 static void
@@ -1637,7 +1700,7 @@ cast_set_status_cb(struct output_session *session, output_status_cb cb)
 static int
 cast_init(void)
 {
-  int mdns_flags;
+  int family;
   int i;
   int ret;
 
@@ -1668,9 +1731,12 @@ cast_init(void)
       goto out_tls_deinit;
     }
 
-  mdns_flags = MDNS_WANT_V4 | MDNS_WANT_V6 | MDNS_WANT_V6LL;
+  if (cfg_getbool(cfg_getsec(cfg, "general"), "ipv6"))
+    family = AF_UNSPEC;
+  else
+    family = AF_INET;
 
-  ret = mdns_browse("_googlecast._tcp", mdns_flags, cast_device_cb);
+  ret = mdns_browse("_googlecast._tcp", family, cast_device_cb);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_CAST, "Could not add mDNS browser for Chromecast devices\n");
