@@ -2,12 +2,8 @@
 #ifndef __PIPE_H__
 #define __PIPE_H__
 
+#include <event2/buffer.h>
 #include "db.h"
-#ifdef HAVE_LIBEVENT2
-# include <event2/buffer.h>
-#else
-# include <event.h>
-#endif
 
 int
 pipe_setup(struct media_file_info *mfi);
