@@ -492,7 +492,7 @@ pairing_request_cb(struct evhttp_request *req, void *arg)
 
   ri->pi.guid = strdup(guid);
 
-  DPRINTF(E_INFO, L_REMOTE, "Pairing succeeded with Remote '%s' (id %s), GUID: %s\n", ri->pi.name, ri->pi.remote_id, guid);
+  DPRINTF(E_LOG, L_REMOTE, "Pairing succeeded with Remote '%s' (id %s), GUID: %s\n", ri->pi.name, ri->pi.remote_id, guid);
 
   ret = db_pairing_add(&ri->pi);
   if (ret < 0)
@@ -935,7 +935,8 @@ remote_pairing_init(void)
     }
 #endif /* USE_EVENTFD */
 
-  ret = mdns_browse("_touch-remote._tcp", MDNS_WANT_V4, touch_remote_cb);
+  // No ipv6 for remote at the moment
+  ret = mdns_browse("_touch-remote._tcp", AF_INET, touch_remote_cb);
   if (ret < 0)
     {
       DPRINTF(E_FATAL, L_REMOTE, "Could not browse for Remote services\n");
